from .Game import Game


def main():
    game = Game(width=768, height=700, fps=120)
    while game.isRunning:
        game.update()
        game.handle_event()
        game.render()

    game.clean()


if __name__ == "__main__":
    main()



import pygame
import os
from .engine.GameState import GameState
from .engine.GameStateManager import GameStateManager
from .engine.InputManager import InputManager
from .engine.TextManager import TextManager
from .engine.Button import Button
from .engine.Window import Window
from . import PathPlanningSubProcess
from .config import ASSETS_DIR, IMAGE_DIR
import json


class PathPlanningState(GameState):
    def __init__(self):
        self.buttons = []

        # Button: Go back to the previous state
        back_button = Button(x=300, y=550, width=150, height=60)
        back_button.set_title("Back")
        back_button.on_click(lambda: GameStateManager().pop_state())
        back_button.set_border(2)
        self.buttons.append(back_button)

        with open(os.path.join(ASSETS_DIR, "template.txt"), "r") as file:
            s = file.read()
        templates = json.loads(s)

        for index, template in enumerate(templates):
            n_rescue_team, n_victim = template["type"]
            info = template["info"]

            column = index % 2
            row = index // 2
            width = 300
            height = 80
            pad = 20
            x = 80
            y = 200
            button = Button(
                x=x + column * (width + pad),
                y=y + row * (height + pad),
                width=width,
                height=80,
            )
            button.set_title(f"{n_rescue_team} Rescue Teams, {n_victim} Victims")
            button.set_font_size(30)
            button.set_border(2)

            def create_on_click(info):
                return lambda: PathPlanningSubProcess.run(
                    image_link=os.path.join(IMAGE_DIR, "map1.jpeg"),
                    victim_position=info["victim_positions"],
                    fatals=info["fatals"],
                    victim_needs=info["victim_needs"],
                    rescue_position=info["rescue_positions"],
                    rescue_resources=info["rescue_resources"],
                    assembly_area=[500, 20],
                )

            button.on_click(create_on_click(info))
            self.buttons.append(button)

    def update(self):
        for button in self.buttons:
            button.update()

    def handle_events(self):
        from .PlayState import PlayState

        if InputManager().is_key_down(pygame.K_RETURN):
            GameStateManager().switch_state(PlayState())

    def render(self, window: Window):
        window.fill("white")
        TextManager().print(
            window=window,
            text="UAV4Res: Path Planning Visualizer",
            position=(window.width / 2, 100),
            color="black",
            font_size=50,
        )

        for button in self.buttons:
            button.draw(window)

    def clean(self):
        pass



import pygame
from .engine.GameState import GameState
from .engine.GameStateManager import GameStateManager
from .engine.InputManager import InputManager
from .engine.TextManager import TextManager
from .engine.TextureManager import TextureManager
from .engine.Button import Button
from .engine.Window import Window


class InfoState(GameState):
    def __init__(self):
        self.buttons = []

        from .MenuState import MenuState

        button = Button(x=280, y=500, width=200, height=65)
        button.set_title("Back")
        button.set_border(2)
        button.on_click(lambda: GameStateManager().switch_state(MenuState()))
        self.buttons.append(button)

    def update(self):
        for button in self.buttons:
            button.update()

    def handle_events(self):
        from PlayState import PlayState

        if InputManager().is_key_down(pygame.K_RETURN):
            GameStateManager().switch_state(PlayState())

    def render(self, window: Window):
        window.fill("white")
        TextManager().print(
            text="Fluffy (UET-VNU): UAV4Res",
            position=(50 + window.width / 2, 80),
            color="black",
            font_size=50,
        )

        TextureManager().draw_texture(
            name="uet_logo",
            position=(40, 30),
            scale=(100, 100),
        )

        TextManager().print(
            text="Project from team Fluffy UET-VNU, Project from team Fluffy UET-VNU, Project from team Fluffy UET-VNU",
            position=(window.width / 2, 300),
            color="black",
            font_size=30,
            max_width=500,
        )

        for button in self.buttons:
            button.draw(window)

    def clean(self):
        pass



import pygame
from .engine.GameState import GameState
from .engine.GameStateManager import GameStateManager
from .engine.InputManager import InputManager
from .engine.Window import Window


class PlayState(GameState):
    def update(self):
        pass

    def handle_events(self):
        from .MenuState import MenuState

        if InputManager().is_key_down(pygame.K_ESCAPE):
            GameStateManager().switch_state(MenuState())

    def render(self, window: Window):
        window.fill((0, 255, 0))

    def clean(self):
        pass



from .engine.GameState import GameState
from .engine.GameStateManager import GameStateManager
from .engine.TextManager import TextManager
from .engine.TextureManager import TextureManager
from .engine.Button import Button
from .engine.Window import Window


class MenuState(GameState):
    def __init__(self):
        self.buttons = []
        from .MapBuildingState import MapBuildingState

        button = Button(x=230, y=300, width=300, height=100)
        button.set_title("Start Demo")
        button.set_border(2)
        button.set_font_size(40)
        button.on_click(lambda: GameStateManager().push_state(MapBuildingState()))
        self.buttons.append(button)

    def update(self):
        for button in self.buttons:
            button.update()

    def handle_events(self):
        pass

    def render(self, window: Window):
        window.fill("white")
        TextManager().print(
            window = window,
            text="UAV4Res",
            position=(50 + window.width / 2, 180),
            color="black",
            font_size=50,
        )

        TextureManager().draw_texture(
            name="uet_logo",
            window = window,
            position=(50, 130),
            scale=(120, 120),
        )

        for button in self.buttons:
            button.draw(window)

    def clean(self):
        pass



from TextureManager import TextureManager
from TextManager import TextManager


class RescueArea():
    def __init__(self, position,  number_of_victim: int, important_score: int):
        self.important_score = important_score
        self.position = position
        self.number_of_victim = number_of_victim

    def update(self):
        pass

    def handle_events(self):
        pass

    def draw(self):
        TextureManager().draw_texture(
            name="boat", position=self.position, scale=(30, 30)
        )
        TextManager().print(
            text=f"important_score: {self.important_score},victims: {self.number_of_victim}",
            position=self.position,
            color="black",
            font_size=20,
        )

    def clean(self):
        pass



import pygame
import cv2
import os
import numpy as np
from .engine.GameState import GameState
from .engine.GameStateManager import GameStateManager
from .engine.InputManager import InputManager
from .engine.TextManager import TextManager
from .engine.Button import Button
from .engine.TextureManager import TextureManager
from .engine.Window import Window
from .config import IMAGE_DIR


class MapBuildingState(GameState):
    def __init__(self):
        self.image = cv2.imread(
            os.path.join(IMAGE_DIR, "map1.jpeg")
        )  # Load the default image (replace with your map)
        self.result_image = None  # Will hold the processed map
        self.buttons = []

        # Load a texture for button background

        # Button: Go back to the previous state
        back_button = Button(x=100, y=500, width=150, height=60)
        back_button.set_title("Back")
        back_button.on_click(lambda: GameStateManager().pop_state())
        back_button.set_border(2)
        self.buttons.append(back_button)

        # Button: Navigate to Path Planning State
        path_planning_button = Button(x=500, y=500, width=150, height=60)
        path_planning_button.set_title("Path Planning")
        path_planning_button.disable()
        path_planning_button.set_border(2)
        from .PathPlanningState import PathPlanningState

        # Button: Process map (Next step)
        next_button = Button(x=300, y=500, width=150, height=60)
        next_button.set_title("Build Map")
        next_button.on_click(
            lambda: (self.process_map(), path_planning_button.enable())
        )
        next_button.set_border(2)
        self.buttons.append(next_button)

        path_planning_button.on_click(
            lambda: GameStateManager().push_state(PathPlanningState())
        )
        self.buttons.append(path_planning_button)

    def process_map(self):
        if self.image is not None:
            self.result_image = self.getGroundMap()

    def getGroundMap(self):
        blurred_image = cv2.GaussianBlur(self.image, (5, 5), 0)
        gray_image = cv2.cvtColor(blurred_image, cv2.COLOR_BGR2GRAY)
        edges = cv2.Canny(gray_image, threshold1=0, threshold2=20)
        contours, _ = cv2.findContours(
            edges, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
        )
        contour_map = np.zeros_like(gray_image)
        cv2.drawContours(contour_map, contours, -1, (255), 1)  # Contour thickness = 1
        kernel = np.ones((5, 5), np.uint8)
        morph_image = cv2.morphologyEx(contour_map, cv2.MORPH_CLOSE, kernel)
        groundMapModel = cv2.bitwise_not(morph_image)
        return groundMapModel

    def update(self):
        for button in self.buttons:
            button.update()

    def handle_events(self):
        from .PlayState import PlayState

        if InputManager().is_key_down(pygame.K_RETURN):
            GameStateManager().switch_state(PlayState())

    def render(self, window: Window):
        window.fill("white")
        TextManager().print(
            window=window,
            text="UAV4Res: Map Building",
            position=(window.width / 2, 100),
            color="black",
            font_size=50,
        )

        TextureManager().draw_texture(
            window=window,
            name="arrow",
            position=(310, 240),
            scale=(180, 120),
        )

        TextManager().print(
            window=window,
            text="Segmentation",
            position=(window.width / 2, 280),
            color="black",
            font_size=20,
        )

        # Render buttons
        for button in self.buttons:
            button.draw(window)

        # Render original image
        if self.image is not None:
            original_surface = pygame.surfarray.make_surface(
                cv2.cvtColor(self.image, cv2.COLOR_BGR2RGB)
            )
            window.draw_image(
                texture=original_surface, position=(100, 200), scale=(200, 200), rotation=-90
            )  # Adjust position as needed

        # Render processed map if available
        if self.result_image is not None:
            result_surface = pygame.surfarray.make_surface(
                cv2.cvtColor(self.result_image, cv2.COLOR_GRAY2RGB)
            )
            window.draw_image(
                texture=result_surface, position=(500, 200), scale=(200, 200), rotation=-90
            )  # Adjust position as needed

    def clean(self):
        pass



import os
from .config import IMAGE_DIR
import pygame
from .engine.Singleton import Singleton
from .engine.GameStateManager import GameStateManager
from .engine.InputManager import InputManager
from .engine.Window import Window
from .MenuState import MenuState
from .engine.TextureManager import TextureManager


@Singleton
class Game:
    def __init__(self, width, height, fps):
        pygame.init()
        self.isRunning = True
        self.FPS = fps
        self.window = Window(width, height, self.FPS)
        GameStateManager().push_state(MenuState())
        self.loadTexture()

    def loadTexture(self):
        TextureManager().load_texture("boat", os.path.join(IMAGE_DIR, "boat.png"))
        TextureManager().load_texture(
            "uet_logo", os.path.join(IMAGE_DIR, "uet_logo.png")
        )
        TextureManager().load_texture("arrow", os.path.join(IMAGE_DIR, "arrow.png"))
        TextureManager().load_texture("map1", os.path.join(IMAGE_DIR, "map1.jpeg"))

    def update(self):
        GameStateManager().update()

    def handle_event(self):
        if InputManager().is_quit():
            self.quit()

        self.window.handle_FPS()
        InputManager().update()

        GameStateManager().handle_events()

    def render(self):
        self.window.fill("white")
        GameStateManager().render(self.window)
        pygame.display.flip()

    def quit(self):
        self.isRunning = False

    def clean(self):
        GameStateManager().clean()
        pygame.quit()



from .engine.TextureManager import TextureManager
from .engine.TextManager import TextManager


class RescueTeam:
    def __init__(self, position, number_of_seats: int):
        self.position = position
        self.number_of_seats = number_of_seats
        self.speed = 2

    def update(self):
        pass

    def handle_events(self):
        pass

    def draw(self):
        TextureManager().draw_texture(
            name="boat", position=self.position, scale=(30, 30)
        )
        TextManager().print(
            text=f"seats: {self.number_of_seats}",
            position=self.position,
            color="black",
            font_size=20,
        )

    def clean(self):
        pass



import setuptools

LONG_DESC = open("README.md").read()
VERSION = "0.1.0"
DOWNLOAD = "https://github.com/UAV4Res/uav4res-visualizer/archive/%s.tar.gz" % VERSION

setuptools.setup(
    name="uav4res-visualizer",
    version=VERSION,
    author="UAV4Res",
    author_email="nhphuong.code@gmail.com",
    description="uav4res-visualizer",
    long_description_content_type="text/markdown",
    long_description=LONG_DESC,
    keywords="uav4res-visualizer",
    license="MIT",
    url="https://github.com/UAV4Res/uav4res-visualizer",
    download_url=DOWNLOAD,
    classifiers=[
        "Environment :: X11 Applications",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    packages=setuptools.find_packages(),
    entry_points={
        "console_scripts": ["uav4res_visualizer=uav4res_visualizer.__main__:main"]
    },
    python_requires=">=3.5",
    install_requires=[
        "pygame>=2.4.0",
        "numpy",
        "opencv-python",
        "Pillow",
    ],
    package_data={"uav4res-visualizer": ["engine/*"]},
    include_package_data=True,
    zip_safe=False,
)



template1 = {
    "image": "map1.png",
    "rescue_team": [
        {"position": (0, 0), "seats": 2},
        {"position": (50, 100), "seats": 2},
        {"position": (100, 200), "seats": 2},
        {"position": (200, 300), "seats": 2},
    ],
    "rescue_area": [
        {"position": (0, 0), "number_of_victims": 3, "important_score": 3},
        {"position": (0, 0), "number_of_victims": 3, "important_score": 3},
        {"position": (0, 0), "number_of_victims": 3, "important_score": 3},
        {"position": (0, 0), "number_of_victims": 3, "important_score": 3},
        {"position": (0, 0), "number_of_victims": 3, "important_score": 3},
    ],
}



import pygame
import numpy
import cv2
from PIL import Image
from . import helper
from multiprocessing import Process, Queue
import time
import sys

#WIDTH, HEIGHT = 1280, 720
WIDTH, HEIGHT = 640, 480

# fatals = [7, 5, 9, 5, 1]
# rescue_resources = [5, 2, 3, 4, 1]
# victim_needs = [2, 4, 5, 3, 1]
# is_running_algo = False

# fatals = [7, 5, 9, 5, 1]
# rescue_resources = [5, 2, 3, 4, 1]
# victim_needs = [2, 4, 5, 3, 1]

# fatals = [7, 1]
# rescue_resources = [5, 9]
# victim_needs = [4, 6]

# fatals = [7]
# rescue_resources = [5]
# victim_needs = [4]

scale = 1.5

class PathPlanningSubProcess():
    def __init__(self, image_link, victim_position, fatals, victim_needs, rescue_position, rescue_resources, assembly_area):
        self.screen = pygame.display.set_mode((WIDTH * scale, HEIGHT * scale))
        pygame.display.set_caption("Visualize");
        self.font = pygame.font.Font('freesansbold.ttf', 10)

        # Load background
        self.background = numpy.array(Image.open(image_link));
        self.background = cv2.resize(self.background, (WIDTH, HEIGHT))
        self.background = numpy.transpose(self.background, (1, 0, 2))
        self.background_original = numpy.array(self.background)
        self.background_with_paths = numpy.array(self.background)

        self.screen.fill((0, 0, 0))
        self.screen.blit(pygame.surfarray.make_surface(self.background), (0, 0))

        self.fatals = fatals
        self.rescue_resources = rescue_resources
        self.return_rescure_resources = []
        self.victim_needs = victim_needs

        # Rescue teams and victims position
        self.res = rescue_position
        self.vic = victim_position
        self.return_res = []
        self.assembly_area = assembly_area

        

        # Resulting path    
        self.paths = [None for i in self.res]
        self.return_res_paths = [None for i in self.return_res]
        # Time controller
        self.running_time_start = time.time()
        self.time_unit = 0;
        self.start_time_window = time.time();
        self.time_window = 0.02;
        self.start_draw_time = -1;
        self.start_draw_return_time = -1;
        self.stop_process = True
        self.stop_process_return = False
        self.start = True
    def draw_box(self, box, colour, thickness):
        boxX, boxY = box
        pygame.draw.rect(self.screen, colour, (boxX - thickness, boxY - thickness, thickness * 2, thickness * 2));
    
    def draw_text(self, box, text, color = (0, 255, 255)):
        boxX, boxY = box
        text = self.font.render(text, True, color)
        textRect = text.get_rect()
        textRect.bottomleft = (boxX + 10, boxY - 10)
        self.screen.blit(text, textRect)

    def draw_paths(self):
        for path in self.paths:
            if path is not None:
                for cell in path:
                    self.draw_box([cell[0] * scale, cell[1] * scale], colour = (255, 0, 255), thickness = 2)
        for path in self.return_res_paths:
            if path is not None:
                for cell in path:
                    self.draw_box([cell[0] * scale, cell[1] * scale], colour = (255, 0, 255), thickness = 2)

    def draw_res_vic(self):
        self.draw_box([self.assembly_area[0][0] * scale, self.assembly_area[0][1] * scale], colour = (255, 0, 0), thickness = 5)
        for index_single_vic, single_vic in enumerate(self.vic):
            self.draw_box([single_vic[0] * scale, single_vic[1] * scale], colour = (0, 0, 255), thickness = 5) 
            self.draw_text([single_vic[0] * scale, single_vic[1] * scale], f"fatals: {self.fatals[index_single_vic]}")
            self.draw_text([single_vic[0] * scale, (single_vic[1] + 10) * scale], f"victim needs: {self.victim_needs[index_single_vic]}")
        for index_single_res, single_res in enumerate(self.res):
            self.draw_box([single_res[0]* scale, single_res[1]* scale], colour = (0, 255, 0), thickness = 5)
            self.draw_text([single_res[0]* scale, single_res[1]* scale], f"rescue resources: {self.rescue_resources[index_single_res]}")
        for index_single_return_res, single_return_res in enumerate(self.return_res):
            self.draw_box([single_return_res[0]* scale, single_return_res[1]* scale], colour = (0, 255, 0), thickness = 5)
            
    def main(self, queue):
        self.screen.blit(pygame.surfarray.make_surface(cv2.resize(self.background, (int(HEIGHT * scale), int(WIDTH * scale)))), (0, 0))
        self.draw_paths()   
        self.draw_res_vic();
        self.draw_text([5* scale, 30* scale], f'running time: {time.time() - self.running_time_start:.2f}', (255, 255, 255))
        if self.stop_process == True or self.stop_process_return == True:
            dot = (int(time.time() * 4) % 4) * '.'
            self.draw_text([5* scale, 40* scale], f'calculating ' + dot, (255, 255, 255))
        mouse_x, mouse_y = pygame.mouse.get_pos()
        if time.time() - self.start_time_window > self.time_window and self.stop_process == False and self.stop_process_return == False:
            self.time_unit = self.time_unit + 1;
            self.start_time_window = time.time()

            if not self.start_draw_return_time == -1:
                if self.stop_process_return == False:
                    return_index = self.time_unit - self.start_draw_return_time

                    for single_return_res_index in range(len(self.return_res)):
                        if self.return_res_paths[single_return_res_index] is not None:
                            if return_index < len(self.return_res_paths[single_return_res_index]):
                                self.return_res[single_return_res_index] = list(self.return_res_paths[single_return_res_index][return_index])
                            else:
                                self.res.append(self.return_res[single_return_res_index])
                                self.rescue_resources.append(self.return_rescure_resources[single_return_res_index])

                                del self.return_res[single_return_res_index]
                                del self.return_rescure_resources[single_return_res_index]
                                del self.return_res_paths[single_return_res_index]

                                queue.put([True, 
                                    numpy.array(self.background),
                                    self.res,
                                    self.vic,
                                    self.fatals,
                                    self.rescue_resources,
                                    self.victim_needs
                                ])
                                print(self.res, self.vic, self.fatals, self.rescue_resources, self.victim_needs)
                                self.start_draw_time = -1
                                self.stop_process = True
                                break;

            if not self.start_draw_time == -1:
                if self.stop_process == False:
                    index = self.time_unit - self.start_draw_time
                    for single_res_index in range(len(self.res)):
                        if self.paths[single_res_index] is not None: 
                            if index < len(self.paths[single_res_index]):
                                self.res[single_res_index] = list(self.paths[single_res_index][index])
                            else:
                                print(self.res, self.vic, self.fatals, self.rescue_resources, self.victim_needs)
                                single_vic_index = self.vic.index(self.res[single_res_index])

                                # Del rescure team
                                if self.victim_needs[single_vic_index] > self.rescue_resources[single_res_index]:
                                    self.victim_needs[single_vic_index] -= self.rescue_resources[single_res_index];

                                    self.return_res.append(self.res[single_res_index])
                                    self.return_rescure_resources.append(self.rescue_resources[single_res_index])
                                    
                                    del self.res[single_res_index]
                                    del self.rescue_resources[single_res_index]
                                    del self.paths[single_res_index]
                                    
                                    queue.put([True, 
                                        numpy.array(self.background),
                                        self.return_res,
                                        self.assembly_area
                                    ])
                                    print(self.return_res, self.assembly_area)
                                    self.start_draw_return_time = -1;
                                    self.stop_process_return = True

                                # Del victim
                                elif self.victim_needs[single_vic_index] < self.rescue_resources[single_res_index]:
                                    self.rescue_resources[single_res_index] -= self.victim_needs[single_vic_index]
                                    del self.victim_needs[single_vic_index]
                                    del self.fatals[single_vic_index]
                                    del self.vic[single_vic_index]
                                else:

                                    self.return_res.append(self.res[single_res_index])
                                    self.return_rescure_resources.append(self.rescue_resources[single_res_index])
                                    
                                    del self.res[single_res_index]
                                    del self.paths[single_res_index]
                                    del self.vic[single_vic_index]
                                    del self.fatals[single_vic_index]
                                    del self.rescue_resources[single_res_index]
                                    del self.victim_needs[single_vic_index]

                                    queue.put([True, 
                                        numpy.array(self.background),
                                        self.return_res,
                                        self.assembly_area
                                    ])

                                    print(self.return_res, self.assembly_area)
                                    self.start_draw_return_time = -1;
                                    self.stop_process_return = True

                                #print(self.res, self.vic, self.fatals, self.rescue_resources, self.victim_needs)
                                queue.put([True, 
                                        numpy.array(self.background),
                                        self.res,
                                        self.vic,
                                        self.fatals,
                                        self.rescue_resources,
                                        self.victim_needs
                                ])
                                self.start_draw_time = -1
                                self.stop_process = True
                                break;
                        elif self.res[single_res_index] != self.assembly_area[0]:
                            self.return_res.append(self.res[single_res_index])
                            self.return_rescure_resources.append(self.rescue_resources[single_res_index])
                            
                            del self.res[single_res_index]
                            del self.rescue_resources[single_res_index]
                            del self.paths[single_res_index]
                            
                            queue.put([True, 
                                numpy.array(self.background),
                                self.return_res,
                                self.assembly_area
                            ])
                            print(self.return_res, self.assembly_area)
                            self.start_draw_return_time = -1;
                            self.stop_process_return = True
                            break;

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                sys.exit()
            elif event.type == pygame.KEYDOWN:
                key = chr(event.key)
                if key == 'q':
                    pygame.quit()
                    sys.exit()
            #     if key == 's':
            #         self.res.append([mouse_x, mouse_y]);
            #         #print(mouse_x, mouse_y)
            #         self.paths.append(None)
            #     if key == 'e':
            #         self.vic.append([mouse_x, mouse_y])
            #         print(mouse_x, mouse_y)
            #     if key == 'r':
            #         queue.put([True, 
            #                 numpy.array(self.background),
            #                 self.res,
            #                 self.vic,
            #                 self.fatals,
            #                 self.rescue_resources,
            #                 self.victim_needs
            #                    ])
            #         self.start_draw_time = -1
        if self.start == True:
            queue.put([True, 
                    numpy.array(self.background),
                    self.res,
                    self.vic,
                    self.fatals,
                    self.rescue_resources,
                    self.victim_needs
                        ])
            self.start_draw_time = -1
            self.start = False
        pygame.display.update()

def main_loop(queue, image_link, victim_position, fatals, victim_needs, rescue_position, rescue_resources, assembly_area):
    pygame.init()

    algo_thres = Process(target=algo, args=(queue, ), daemon=True);
    algo_thres.start()
    
    gui = PathPlanningSubProcess(image_link, victim_position, fatals, victim_needs, rescue_position, rescue_resources, assembly_area)
    
    while True:
        #print("->")
        try:
            receiver = queue.get_nowait()
            #print("main loop keep loop !!!1")
        except:
            gui.main(queue)
            continue;
        if len(receiver) == 7:
            #print("main loop keep loop !!!2")
            queue.put(receiver)
            continue
        elif len(receiver)  == 4:
            queue.put(receiver)
            #print("main loop keep loop !!!3")
            continue
        elif len(receiver) == 2:
            paths = receiver[1]
            gui.paths = paths
            gui.start_draw_time = gui.time_unit
            gui.stop_process = False
            gui.start_time_window = time.time()
        elif len(receiver) == 3:
            return_res_paths = receiver[1]
            gui.return_res_paths = return_res_paths
            gui.start_draw_return_time = gui.time_unit
            gui.stop_process_return = False
            gui.start_time_window = time.time()
        gui.main(queue)
    
def algo(queue):
    is_running_algo = False
    while True:
        try:
            receiver = queue.get_nowait()
            #print("algo have received!!!")
        except:
            continue;

        if len(receiver) == 7:
            is_running_algo, background, res, vic, fatals, rescue_resources, victim_needs = receiver
            print(res, vic, fatals, rescue_resources, victim_needs)
        elif len(receiver) == 4:
            is_running_algo, background, return_res, assembly_area = receiver
            print(return_res, assembly_area)
        else:
            queue.put(receiver)
            continue

        if is_running_algo:
            # print(background.shape)
            groundMapModel = helper.buildMap(background)
            #print(groundMapModel)
            groundMapModel = numpy.repeat(groundMapModel.reshape(groundMapModel.shape[0], groundMapModel.shape[1], 1), 3, axis=2)
            #print(groundMapModel)
            groundMapModel = helper.turn_image_to_binary(groundMapModel, 1)
            # #print(groundMapModel)
            print("start")
            if len(receiver) == 7:
                res = numpy.array(res);
                vic = numpy.array(vic);

                # res = numpy.concatenate([res[:, 1:], res[:, :1]], axis = 1).tolist();
                # vic = numpy.concatenate([vic[:, 1:], vic[:, :1]], axis = 1).tolist();
                #print(res.shape, vic.shape)

                time1 = time.time()
                paths, _ = helper.solve_for_paths(groundMapModel, res, vic, fatals, rescue_resources, victim_needs);
                time2 = time.time()
                print(time2 - time1)
                print("end")
                queue.put([False, paths])
            if len(receiver) == 4:
                return_res = numpy.array(return_res)
                assembly_area = numpy.array(assembly_area)
                print(return_res, assembly_area)

                paths = helper.paths_for_return(groundMapModel, return_res, assembly_area)
                queue.put([False, paths, -1])
                print("end")
                pass
            is_running_algo = False
            #print("ye")

def run(image_link, victim_position, fatals, victim_needs, rescue_position, rescue_resources, assembly_area):
    """
        Run finding path algorithm
        Args:
            image_link: Link to the background image
            victim_position: Position of victims, data type: 2d list, shape (number of victims, position)
            fatals: Measure how bad the victim get injured, data type: 1d list, shape (number of victims)
            victim_needs: Measure how many rescue resources the victim need, data type: 1d list, shape (number of victims)
            rescue_position: Position of rescue teams, data type: 2d list, shape (number of rescue teams, position)
            rescue_resources: Measure how many rescue resources the rescue team have, data type: 1d list, shape (number of rescue teams)
            assembly_area: Position of the area which the rescue team must bring to when there are no space left in the boat: 1d list, shape (position)
    """
    queue = Queue()
    main_loop(queue, image_link, victim_position, fatals, victim_needs, rescue_position, rescue_resources, [assembly_area])

#if __name__ == '__main__':
    # fatals = [7, 5, 9, 5, 1]
    # rescue_resources = [5, 2, 3, 4, 1]
    # victim_needs = [2, 4, 5, 3, 1]
    # is_running_algo = False
    #run("test.jpg", [[20, 20], [60, 60]], [7, 1], [4, 6], [[100, 50], [400, 400]], [2, 2], [400, 20])
    #run("test.jpg", [], [7, 5, 9], [2, 4, 5], [], [5, 2], [505, 20])




import random
import numpy as np
from PIL import Image
import matplotlib.pyplot as plt
import heapq
import cv2

def turn_image_to_binary(image, threshold):
    image = np.mean(image, axis = 2);
    image[image < threshold] = 0;
    image[image >= threshold] = 1;
    return image

def load_PIL_image(path):
    image = np.array(Image.open(path))[:, :, :3]
    return image

def L1(p1, p2):
    return abs(p1[0] - p2[0]) + abs(p1[1] - p2[1])
    #return np.sqrt((p1[0] - p2[0])**2 + (p1[1] - p2[1])**2);

WALL = 0;

def BFS(image, start):
    """
        image: (W, H) (row, col)
        start: (x, y)
        finish: (x, y)
    """
    image = np.array(image);  

    recent = start
    recent_cost = 0
    potential_nodes = []
    previous_move = [[None for i in range(image.shape[1])] for i in range(image.shape[0])]
    cost_matrix = [[1e9 for i in range(image.shape[1])] for i in range(image.shape[0])]
    visited = np.full(image.shape, False)

    def valid_move(position):
        for i in range(2):
            if position[i] >= image.shape[i] or position[i] < 0:
                return False;

        if image[position] == WALL or visited[position] == True:
            return False;
        return True

    ### BFS algorithm ###
    visited[start] = True
    heapq.heappush(potential_nodes, (0, recent));
    while(len(potential_nodes) != 0):
        recent_cost, recent = heapq.heappop(potential_nodes)
        cost_matrix[recent[0]][recent[1]] = min(cost_matrix[recent[0]][recent[1]], recent_cost)

        next_moves = [(recent[0] - 1 + i, recent[1] - 1 + j) for i in range(3) for j in range(3) if (i != 1 or j != 1)]

        #print(next_moves)
        #break;

        # next_moves = [
        #     (recent[0], recent[1] + 1),
        #     (recent[0] + 1, recent[1]),
        #     (recent[0], recent[1] - 1),
        #     (recent[0] - 1, recent[1])
        # ]

        for next_move in next_moves:
            if valid_move(next_move):
                if recent[0] != next_move[0] and recent[1] != next_move[1]:
                    if image[next_move[0], recent[1]] == WALL or image[recent[0], next_move[1]] == WALL:
                        continue
                move_cost = (np.sqrt(2) if recent[0] != next_move[0] and recent[1] != next_move[1] else 1)
                heapq.heappush(potential_nodes, (recent_cost + move_cost, next_move))
                previous_move[next_move[0]][next_move[1]] = recent;
                visited[next_move] = True

    # ### Trace Back ###
    # while(recent != start):
    #     path = [recent] + path;
    #     recent = previous_move[recent[0]][recent[1]]
    # path = [recent] + path

    return cost_matrix, previous_move, visited;

def a_aristek(image, start, finish, heuristic_func):
    """
        image: (W, H) (row, col)
        start: (x, y)
        finish: (x, y)
    """
    image = np.array(image);  

    recent = start
    recent_cost = 0
    potential_nodes = []
    previous_move = [[None for i in range(image.shape[1])] for i in range(image.shape[0])]
    #print(image.shape)
    visited = np.full(image.shape, False)

    def valid_move(position):
        for i in range(2):
            if position[i] >= image.shape[i] or position[i] < 0:
                return False;

        if image[position] == WALL or visited[position] == True:
            return False;
        return True

    ### A* algorithm ###
    visited[start] = True
    heapq.heappush(potential_nodes, (0, 0, recent));
    while(recent != finish and len(potential_nodes) != 0):
        heuristic_cost, recent_cost, recent = heapq.heappop(potential_nodes)

        next_moves = [(recent[0] - 1 + i, recent[1] - 1 + j) for i in range(3) for j in range(3) if (i != 1 or j != 1)]

        #print(next_moves)
        #break;

        # next_moves = [
        #     (recent[0], recent[1] + 1),
        #     (recent[0] + 1, recent[1]),
        #     (recent[0], recent[1] - 1),
        #     (recent[0] - 1, recent[1])
        # ]

        for next_move in next_moves:
            if valid_move(next_move):
                if recent[0] != next_move[0] and recent[1] != next_move[1]:
                    if image[next_move[0], recent[1]] == WALL or image[recent[0], next_move[1]] == WALL:
                        continue
                move_cost = (np.sqrt(2) if recent[0] != next_move[0] and recent[1] != next_move[1] else 1)
                heapq.heappush(potential_nodes, (recent_cost + move_cost + heuristic_func(recent, finish), recent_cost + move_cost, next_move));
                previous_move[next_move[0]][next_move[1]] = recent;
                visited[next_move] = True

    path = []
    if recent != finish:
        return 1e9, [], visited

    ### Trace Back ###
    while(recent != start):
        path = [recent] + path;
        recent = previous_move[recent[0]][recent[1]]
    path = [recent] + path

    return recent_cost, path, visited;

def theta_aristek(image, start, finish, heuristic):
    parent = np.array([[None for i in range(image.shape[1])] for i in range(image.shape[0])])
    gScore = np.array([[1e9 for i in range(image.shape[1])] for i in range(image.shape[0])])
    
    class custom_heap:
        def __init__(self):
            self.list = []
        def __contains__(self, item):
            return (item in self.list)
        def pop(self):
            return heapq.heappop(self.list)[1]
        def insert(self, node, distance):
            heapq.heappush(self.list, (distance, node))
        def empty(self):
            return (len(self.list) == 0)
    
    open = custom_heap()

    def c(p1, p2):
        return np.sqrt((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2)

    def main(start, goal):
        # This main loop is the same as A*
        gScore[start] = 0
        parent[start] = start
        # Initializing open and closed sets. The open set is initialized 
        # with the start node and an initial cost
        
        open.insert(start, gScore[start] + heuristic(start, goal))
        # gScore[node] is the current shortest distance from the start node to node
        # heuristic(node) is the estimated distance of node from the goal node
        # there are many options for the heuristic such as Euclidean or Manhattan 

        def valid_move(position):
            for i in range(2):
                if position[i] >= image.shape[i] or position[i] < 0:
                    return False;

            if image[position] == WALL or visited[position] == True:
                return False;
            return True

        visited = np.full(image.shape, False)
        visited[start] = True
        while open.empty() == False:
            s = open.pop()
            if s == goal:
                return gScore[s], reconstruct_path(s)
            visited[s] = True
            next_moves = [
                (s[0], s[1] + 1),
                (s[0] + 1, s[1]),
                (s[0], s[1] - 1),
                (s[0] - 1, s[1])
            ]

            for neighbor in next_moves:
                if valid_move(neighbor):
                    if visited[neighbor] == False:
                        update_vertex(s, neighbor, goal)
        return 1e9, None
                
        
    def update_vertex(s, neighbor, goal):
        # This part of the algorithm is the main difference between A* and Theta*
        if line_of_sight(parent[s], neighbor):
            # If there is line-of-sight between parent[s] and neighbor
            # then ignore s and use the path from parent[s] to neighbor 
            if gScore[parent[s]] + c(parent[s], neighbor) < gScore[neighbor]:
                # c(s, neighbor) is the Euclidean distance from s to neighbor
                gScore[neighbor] = gScore[parent[s]] + c(parent[s], neighbor)
                parent[neighbor] = parent[s]
                open.insert(neighbor, gScore[neighbor] + heuristic(neighbor, goal))
        else:
            # If the length of the path from start to s and from s to 
            # neighbor is shorter than the shortest currently known distance
            # from start to neighbor, then update node with the new distance
            if gScore[s] + c(s, neighbor) < gScore[neighbor]:
                gScore[neighbor] = gScore[s] + c(s, neighbor)
                parent[neighbor] = s
                open.insert(neighbor, gScore[neighbor] + heuristic(neighbor, goal))
        
    def reconstruct_path(s):
        total_path = [s]
        # This will recursively reconstruct the path from the goal node 
        # until the start node is reached
        if parent[s] != s:
            total_path = reconstruct_path(parent[s]) + total_path
        return total_path
        
    def line_of_sight(node1, node2):
        x0 = node1[0];
        y0 = node1[1];
        x1 = node2[0];
        y1 = node2[1];
        dx = abs(x1 - x0);
        dy = -abs(y1 - y0);

        sX = -1;
        sY = -1;
        if(x0 < x1):
            sX = 1;

        if(y0 < y1):
            sY = 1;


        e = dx + dy;
        while(True):
            point = (int(x0), int(y0));
            if(image[point] == WALL):
                return False;

            if(x0 == x1 and y0 == y1):
                return True;

            e2 = 2 * e;
            if(e2 >= dy):
                if(x0 == x1):
                    return True;
                e += dy;
                x0 += sX;

            if(e2 <= dx):
                if(y0 == y1):
                    return True;
                e += dx;
                y0 += sY;
    cost, path = main(start=start, goal=finish)
    real_paths = []
    def draw_bigger(first_axis, second_axis, swap = False):
        # first_axis always longer than second axis
        line_path = []
        for x in range(first_axis[0], first_axis[1], 1 if first_axis[0] < first_axis[1] else -1):
            y = int(second_axis[0] + (x - first_axis[0]) / (first_axis[1] - first_axis[0]) * (second_axis[1] - second_axis[0]))
            if swap:
                line_path.append((y, x));
            else:
                line_path.append((x, y));

        return line_path
    def line(p1, p2):
        if(abs(p1[0] - p2[0]) > abs(p1[1] - p2[1])):
            return draw_bigger((p1[0], p2[0]), (p1[1], p2[1]), swap = False);
        else:
            return draw_bigger((p1[1], p2[1]), (p1[0], p2[0]), swap = True);
    if path != None:        
        for index in range(len(path) - 1):
            pre_point = path[index]
            nex_point = path[index + 1]
            # print("!!!")
            # print(pre_point, nex_point)
            # print(line(pre_point, nex_point))
            real_paths = real_paths + line(pre_point, nex_point)
        real_paths.append(path[len(path) - 1])
    else:
        real_paths = None
    return cost, real_paths, None

def buildMap(image):
  # Step 1: Gaussian Blurring
  blurred_image = cv2.GaussianBlur(image, (5, 5), 0)
  # Step 2: Convert to Grayscale
  gray_image = cv2.cvtColor(blurred_image, cv2.COLOR_BGR2GRAY)
  # Step 3: Canny Edge Detection
  edges = cv2.Canny(gray_image, threshold1=0, threshold2=20)
  # Step 4: Find and Draw Contours
  contours, _ = cv2.findContours(edges, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
  # Create an empty black image for the contour map (black background)
  contour_map = np.zeros_like(gray_image)
  # Draw contours in white (255) on the black background
  cv2.drawContours(contour_map, contours, -1, (255), 1)  # Contour thickness = 1
  # Step 5: Morphological Operations (closing gaps in contours)
  kernel = np.ones((5, 5), np.uint8)
  morph_image = cv2.morphologyEx(contour_map, cv2.MORPH_CLOSE, kernel)
  groundMapModel = cv2.bitwise_not(morph_image)
  # Save or display the final black and white map
  return groundMapModel

def solve_for_path_with_BFS(image, rescue_pos, victim_pos, fatals, rescue_resources, victim_needs):
    rescue_pos = [tuple(single_rescue_pos) for single_rescue_pos in rescue_pos]
    victim_pos = [tuple(single_victim_pos) for single_victim_pos in victim_pos]
    num_of_rescue_teams = len(rescue_pos);
    num_of_victims = len(victim_pos)
    costs_matrix = np.zeros((num_of_rescue_teams, num_of_victims))
    paths_matrix = [[None for i in range(num_of_victims)] for i in range(num_of_rescue_teams)]

    def trace_back(recent, start, previous_move):
        ### Trace Back ###
        path = []
        while(recent != start):
            path = [recent] + path;
            recent = previous_move[recent[0]][recent[1]]
        path = [recent] + path
        
        return path
    
    for i in range(len(rescue_pos)):
        costs_matrix_of_rescure_team, previous_move, _ = BFS(image, rescue_pos[i])
        for j in range(len(victim_pos)):
            costs_matrix[i, j] = costs_matrix_of_rescure_team[victim_pos[j][0]][victim_pos[j][1]]
            paths_matrix[i][j] = trace_back(victim_pos[j], rescue_pos[i], previous_move)
            # if algorithm == "theta_aristek":
            #     costs_matrix[i, j], paths_matrix[i][j], _ = theta_aristek(image, rescue_pos[i], victim_pos[j], L1)
            # else:
            #     costs_matrix[i, j], paths_matrix[i][j], _ = a_aristek(image, rescue_pos[i], victim_pos[j], L1)

    #print(costs_matrix, paths_matrix)
    rescue_remain = [True for i in range(num_of_rescue_teams)]
    rescue_paths = [None for i in range(num_of_rescue_teams)]
    rescue_order = np.argsort(fatals)[::-1]
    for victim_index in rescue_order:
        victim_need = victim_needs[victim_index]
        rescue_cost_order = np.argsort(costs_matrix[:, victim_index].tolist());

        total_resource_for_the_victim = 0;
        for rescue_index in rescue_cost_order:
            if rescue_remain[rescue_index] == True:
                rescue_remain[rescue_index] = False
                rescue_resource = rescue_resources[rescue_index]

                rescue_paths[rescue_index] = paths_matrix[rescue_index][victim_index]

                total_resource_for_the_victim += rescue_resource
                if total_resource_for_the_victim > victim_need:
                    break;

    return rescue_paths, costs_matrix

def solve_for_paths(image, rescue_pos, victim_pos, fatals, rescue_resources, victim_needs, algorithm = "theta_aristek"):
    rescue_pos = [tuple(single_rescue_pos) for single_rescue_pos in rescue_pos]
    victim_pos = [tuple(single_victim_pos) for single_victim_pos in victim_pos]
    num_of_rescue_teams = len(rescue_pos);
    num_of_victims = len(victim_pos)
    costs_matrix = np.zeros((num_of_rescue_teams, num_of_victims))
    paths_matrix = [[None for i in range(num_of_victims)] for i in range(num_of_rescue_teams)]

    for i in range(len(rescue_pos)):
        for j in range(len(victim_pos)):
            if algorithm == "theta_aristek":
                costs_matrix[i, j], paths_matrix[i][j], _ = theta_aristek(image, rescue_pos[i], victim_pos[j], L1)
            else:
                costs_matrix[i, j], paths_matrix[i][j], _ = a_aristek(image, rescue_pos[i], victim_pos[j], L1)
            #print(i, j, costs_matrix[i, j])

    #print(costs_matrix, paths_matrix)
    rescue_remain = [True for i in range(num_of_rescue_teams)]
    rescue_paths = [None for i in range(num_of_rescue_teams)]
    rescue_order = np.argsort(fatals)[::-1]
    for victim_index in rescue_order:
        victim_need = victim_needs[victim_index]
        rescue_cost_order = np.argsort(costs_matrix[:, victim_index].tolist());

        total_resource_for_the_victim = 0;
        for rescue_index in rescue_cost_order:
            if rescue_remain[rescue_index] == True:
                rescue_remain[rescue_index] = False
                rescue_resource = rescue_resources[rescue_index]

                rescue_paths[rescue_index] = paths_matrix[rescue_index][victim_index]

                total_resource_for_the_victim += rescue_resource
                if total_resource_for_the_victim > victim_need:
                    break;

    return rescue_paths, costs_matrix

def paths_for_return(image, rescue_pos, assembly_area):
    rescue_pos = [tuple(single_rescue_pos) for single_rescue_pos in rescue_pos]
    assembly_area = [tuple(assembly_area[0])]
    num_of_rescue_teams = len(rescue_pos);
    paths_matrix = [None for i in range(num_of_rescue_teams)]

    for i in range(len(rescue_pos)):
        _, paths_matrix[i], __ = theta_aristek(image, rescue_pos[i], assembly_area[0], L1)

    return paths_matrix